#!/usr/bin/env python

from __future__ import print_function

"""
A simple client for SignServer

We contact a signing server, and for each file to be signed, send a
digest (actually the hex formatted digest) to be signed.
This limits the traffic over the network, and since the server sends
back a PEM encoded signature, the entire transaction is thus ascii.

We pay attention to the name we were invoked as, and look for
myname.cf to pickup defaults for the url to contact and which
hash to use.

Options:
    -c "config"
        load config from "config"

    -d  bump the debug level

    -u "url"
        connect to the given "url" (host:port)

    -h "hash"
        use "hash" to compute the digest we send to the server.
        The default (sha256) is sufficient for any reasonable
        signature method, though it is perhaps best to use the same
        method as the server.
        So if signing with RSA+SHA1, use sha1 as the digest to send.
        This is especially important if the server will use the digest
        directly.  If we send the wrong input it will return an error.

    -o "ofile"
        controls the file we will save the signature in.
        The default is '.', which means append the extension provided
        by the server to the file that was signed.
        Thus if signing '/some/file' and the server tells us to use
        '.sig' we will save the signature in '/some/file.sig'
        If "ofile" is '-' the signature will be written to stdout.
        
"""

"""
RCSid:
	$Id: sign.py,v 1.37 2014/10/30 23:48:34 sjg Exp $

	@(#) Copyright (c) 2012 Simon J. Gerraty

	This file is provided in the hope that it will
	be of use.  There is absolutely NO WARRANTY.
	Permission to copy, redistribute or otherwise
	use this file is hereby granted provided that 
	the above copyright notice and this notice are
	left intact. 
      
	Please send copies of changes and bug-fixes to:
	sjg@crufty.net

"""

import os.path, socket, sys, time

def basename(path, ext=None):
    b = os.path.basename(path)
    if ext:
        x = b.rfind(ext)
        if x > 0:
            return b[0:x]
    return b

def hash_file(file, hash):
    """hash file using the indicated method
    If file.$hash exists and is up to date,
    we will use its content.
    """
    h = hash()
    hfile = '{}.{}'.format(file, h.name)
    try:
        sth = os.stat(hfile)
        # size of hfile needs to be what we expect
        if sth.st_size == (1 + 2 * h.digest_size):
            # and it needs to be up to date
            stf = os.stat(file)
            if sth.st_mtime >= stf.st_mtime:
                f = open(hfile, 'rb')
                x = f.read()
                f.close()
                return x
    except:
        pass
    f = open(file, 'rb')
    for line in f:
        h.update(line)
    f.close()
    return h.hexdigest()+'\n'

def getConfig(conf, key, default=None):
    """return value of key if present in config, else default"""
    if key in conf:
        return conf[key]
    return default

def loadConfig(name, conf={}):
    """Add key value pairs into conf"""
    cf = open(name, 'r')
    for line in cf:
        if line.startswith('#'):
            continue
        if line.find('=') > 0:
            k,v = line.split('=', 1)
            conf[k.strip()] = v.strip()
    return conf

def url2sock_address(url):
    """parse url and return appropriate socket family and address"""
    if url.startswith('/'):
        sock_family = socket.AF_UNIX
        connect_to = url
    else:
        sock_family = socket.AF_INET
        host = None
        if url.find(':') >= 0:
            (host,port) = url.split(':')
            port = int(port)
        else:
            port = int(url)
        if not host:
            host = 'localhost'

        connect_to = (host,port)
    return (sock_family,connect_to)

def urlconnect(sock_family,connect_to,tries=1):
    for i in range(tries):
        try:
            em = ''
            sock = socket.socket(sock_family, socket.SOCK_STREAM)
            sock.connect(connect_to)
            break
        except Exception as e:
            em = 'ERROR: {0}: {1}: {2}'.format(e.errno, connect_to, e.strerror)
            if isinstance(e, socket.gaierror):
                if not e.errno in [2]:          # most not worth retrying
                    break
            elif isinstance(e, socket.herror):
                if not e.errno in [2]:          # most not worth retrying
                    break
            s = i % 8
            if s > 0:
                time.sleep(s)
            if s == 0 and i + 1 < tries:
                sys.stderr.write(em+'. retrying...\n')
            sock.close()
            sock = None
    if em:
        if sock:
            sock.close()
            sock = None
    return (sock,em)
    
def error(msg, exit=1):
    p = 'ERROR: '
    if msg.startswith('ERROR'):
        p = ''
    sys.stderr.write('{}{}\n'.format(p,msg))
    sys.exit(exit)

if __name__ == '__main__':
    import getopt

    Myname = basename(sys.argv[0], '.py')
    Mydir = os.path.dirname(sys.argv[0])

    conf = {}

    myconf = Myname + '.cf'
    for d in [Mydir, '.']:
        cf = d + '/' + myconf
        if os.path.exists(cf):
            conf = loadConfig(cf, conf)

    debug = 0
    certsFile = None
    tries = 10
    user = None

    # by default we save the sig for file in file.ext
    # using whatever the server tells us for ext
    ofile = '.'

    opts,args = getopt.getopt(sys.argv[1:], 'dc:u:o:h:C:r:x')
    for o,a in opts:
        if o == '-c':
            conf = loadConfig(a, conf)
        elif o == '-C':
            if a == '-':
                certsFile = sys.stdout
            else:
                certsFile = open(a, 'wb')
        elif o == '-d':
            debug += 1
        elif o == '-u':
            conf['url'] = a
        elif o == '-h':
            conf['hash'] = a
        elif o == '-o':
            ofile = a
        elif o == '-r':
            tries = int(a)
        elif o == '-x':
            import pwd
            user = pwd.getpwuid(os.getuid())[0]

    port = getConfig(conf, 'ListenPort', 6161)
    url = getConfig(conf, 'url', 'localhost:{}'.format(port))
    hname = getConfig(conf, 'hash', 'sha256')
    m = __import__('hashlib', fromlist=[hname])
    hash = getattr(m, hname)

    sock_family,connect_to = url2sock_address(url)

    # we don't want to connect to server until we know we have work for it
    sock = None
    sig_ext = None

    if certsFile:
        for i in range(tries):
            try:
                certsFile.seek(0)

                sock,response = urlconnect(sock_family, connect_to, tries)

                if not sock:
                    error(response)
                sock.send('certs:')
                while True:
                    msg = sock.recv(1024)
                    if not msg:
                        break
                    if msg.startswith('ERROR'):
                        error(msg)
                    certsFile.write(msg)
                certsFile.close()
                sock.close()
                sys.exit(0)
            except Exception:
                pass

    if len(args) == 0:
        error('need something to sign')

    for f in args:
        if user:
            msg = 'user={} path={} hash={}'.format(user,f,hash_file(f, hash))
        else:
            msg = hash_file(f, hash)

        for i in range(tries):
            try:
                if not sock:
                    sock,response = urlconnect(sock_family, connect_to, tries)
                if not sock:
                    break
                sock.send(msg)
                response = sock.recv(1024)
                break
            except Exception as e:
                response = 'ERROR: {0}: {1}: {2}'.format(e.errno, url, e.strerror)
                s = i % 8
                if s > 0:
                    time.sleep(s)
                if s == 0 and i + 1 < tries:
                    sys.stderr.write(response+'. retrying...\n')
                sock.close()
                sock = None

        if not response:
            # server may have rejected us due to acl
            response = 'ERROR: no response from: {0}'.format(url)

        if response.startswith('ERROR'):
            error(response)

        if response.startswith('#set:'):
            knobs,response = response.split('\n', 1)
            if debug:
                print(knobs)
            if ofile == '.' and not sig_ext:
                for knob in knobs.split()[1:]:
                    k,v = knob.split('=')
                    if k == 'sig_ext':      # the only one we care about
                        sig_ext = v
                        break
        
        if ofile == '-':
            print(response, end='')
        else:
            if ofile == '.':
                sfile = f + sig_ext
            elif ofile.startswith('.'):
                sfile = f + ofile
            else:
                sfile = ofile
            open(sfile, 'wb').write(response)
